#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include "util.h"
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
//#include "SHMEM.h"

int FSsendlocal(int command, int datalen,const char *fsp,const char *data,int fileno=0);

extern int errno;

void error(char *msg,int pri) {
printf("ERROR: %s\n",msg);
if (pri==ERR_EXIT) exit(1);
}

void error(int pri,char *msg) {
printf("ERROR: %s\n",msg);
if (pri==ERR_EXIT) exit(1);
}

// This will lock a file. If it's already locked it will wait 30 seconds
// then remove the old lock.
int file_lock_wait(FILE *file) {
if (!file) error(ERR_EXIT,"Tried to lock NULL file");

#ifdef SGI
int fdes = file->_file;
#else
int fdes = file->_fileno;
#endif

struct flock fl = { F_WRLCK, SEEK_SET, 0, 0, getpid() };
if (fcntl(fdes,F_SETLKW,&fl)==-1) {
	printf("Error: file locking error (%d)! NFS problem?\n",errno);
	int i;
	for (i=0; i<5; i++) if (fcntl(fdes,F_SETLKW,&fl)!=-1) break; else sleep(1);
	if (i==5) error(ERR_EXIT,"Fatal file locking error");
}

return 0;
}

int file_lock(FILE *file) {
if (!file) error(ERR_EXIT,"Tried to lock NULL file");

#ifdef SGI
int fdes = file->_file;
#else
int fdes = file->_fileno;
#endif
struct flock fl = { F_WRLCK, SEEK_SET, 0, 0, getpid() };

return fcntl(fdes,F_SETLK,&fl);
}

// deprecated, no longer necessary
void file_unlock(FILE *file) {
}

// used for parsing special delimited HTML files for generating
// custom sets of instructions
void addHTMLSection(char *sourcefile,FILE *out,char *secid) {
FILE *in;
int c;
char s[20],*H;
int i;

H=EMDIR();

in=fopen(sourcefile,"r");
if (!in) return;

// rely on buffered IO to make this reasonably efficient
// first find the correct section
s[0]=0;
while (strcmp(s,secid)) {
//	printf("'%s' != '%s'\n",s,secid);
	while ((c=getc(in))!='$' && c!=EOF) ;
	if (c==EOF) return;

	i=0;
	while ((c=getc(in))>' ' && i<10) { s[i]=c; i++; }
	s[i]=0;
}

// dump the section to the output stream
while ((c=getc(in))!='$' && c!=EOF) {
	if (c=='@') {
		c=getc(in);
		if (c=='E') fprintf(out,"%s",H);
		else if (c=='L') {
			char *cw=getcwd(NULL,0);
			if (cw) {
				fprintf(out,"%s",cw);
				free(cw);
			}
		}
		else putc(c,out);
	}
	else putc(c,out);
}

fclose(in);

}

// This returns the path to the EMAN directory
char *EMDIRC = NULL;	//DON'T USE THIS

char *EMDIR() {
if (EMDIRC) return EMDIRC;

if (getenv("EMANDIR")) {
	EMDIRC=(char *)malloc(strlen(getenv("EMANDIR"))+1);
	strcpy(EMDIRC,getenv("EMANDIR"));
}
else {
	if (!getenv("HOME")) { 
		printf("ERROR: Please set environment variable EMANDIR to the location of the 'EMAN' directory\n");
		return("");
	}
	EMDIRC=(char *)malloc(strlen(getenv("HOME"))+6);
	strcpy(EMDIRC,getenv("HOME"));
	strcat(EMDIRC,"/EMAN");
}
return EMDIRC;
}

// All applications should call this before parsing the command line
void appinit(int argc,char *argv[]) {

srandom(time(0));

if (argc>1 && (strcmp(argv[1],"help")==0 || strcmp(argv[1],"-help")==0 ||
strcmp(argv[1],"--help")==0)) {
	FILE *rh;
	char s[256],*t;
	t=strrchr(argv[0],'/');
	if (t) t++; else t=argv[0];

/*	if (rh=DFopenr("remotehelp")) {
		fclose(rh);
		sprintf(s,"netscape -remote 'openURL(http://ncmi.bcm.tmc.edu/~stevel/EMAN/doc/progs/%s.html,new-window)'",t);
	}
	else sprintf(s,"netscape -remote 'openFile(%s/doc/progs/%s.html,new-window)'",EMDIR(),t);

	if (run(s)) {
		printf("Sorry, netscape doesn't appear to be running on your remote display.\nCheck the $DISPLAY variable.\n");
		
	}*/
	sprintf(s,"QHelp %s/doc/progs/%s.html",EMDIR(),t);
	run(s);

	appusage(argv[0]);
	exit(0);
}

#ifdef FILESERVER
FILE *in = DFopenr(".key");
if (!in) {
	in=DFopenw(".key");
	fprintf(in,"%d\n",random());
	fclose(in);
	in=DFopenr(".key");
}
fscanf(in,"%d",&FSKEY);
fclose(in);

for (int i=1; i<argc; i++) {
	if (strncmp("usefs=",argv[i],6)==0) sscanf(&argv[i][6],"%hu,%u",&FSPORT,&FSHOST);
}
#endif



#ifdef EMVERD
printf("EMAN %sdbg.\nRun '%s help' for detailed help.\n",EMVER,argv[0]);
#else
printf("EMAN %s.\nRun '%s help' for detailed help.\n",EMVER,argv[0]);
#endif

}

void appusage(char *argv0) {
char s[256],*t;

t=strrchr(argv0,'/');
if (t) t++; else t=argv0;

sprintf(s,"cat %s/doc/usage/%s.usg",EMDIR(),t);
run(s);
exit(1);
}

int  LOGbegin(int argc,char *argv[],int ppid) {
FILE *out,*in;
time_t tm=time(0);
int ref=getpid(),fl=0;
char *t;
//char *c=(char *)getcwd(NULL,1024);
char *c=getenv("PWD");

if (!c) { c=(char *)getcwd(NULL,1024); fl=1; }

// write the actual log entry
if (strrchr(argv[0],'/')) t=strrchr(argv[0],'/')+1;
else t=argv[0];

char s[2048];	//perhaps dangerous
sprintf(s,"%d\t%d\t%d\t%d\t%s",ref,tm,LOG_BEGIN,ppid?ppid:getppid(),t);
for (int i=1; i<argc; i++) sprintf(s+strlen(s)," %s",argv[i]);
sprintf(s+strlen(s),"\n");

#ifdef FILESERVER
if (FSHOST) { 
	FSsend(FS_TEXT,strlen(s)+1, ".emanlog", s);
	return ref;		// for remote logging, we skip the dirlist 
}
#endif

out=fopen(".emanlog","a");
if (!out) return 0;
file_lock_wait(out);
fprintf(out,"%s",s);
file_unlock(out);
fclose(out);

// add the current directory to the directory list in ~/.eman
in=DFopenr("dirlist");
if (in) {
	char s[1024];
	int f=0;
	while (fscanf(in," %1023s",s)==1) {
		if (strcmp(s,c)==0) { f=1; break;}
	}
	fclose (in);
	if (!f) in=NULL;
}

if (!in) {
	out=DFopena("dirlist");
	fprintf(out,"%s\n",c);
	fclose(out);
}
if (fl) free(c);

return ref;
}

void LOG(int ref,char *file,int mode,char *text) {
FILE *out;
time_t tm=time(0);

if (!text) text="";
if (!file) file="-";

// if we reference a file in a remote directory, put a link in to the
// current directory
if (mode>=10 && mode<=19) {
	if (strrchr(file,'/')) {
		char *f2=(char *)malloc(strlen(file)+8);
		char *c=(char *)getcwd(NULL,1024);
		strcpy(f2,file);
		strcpy(strrchr(f2,'/')+1,".emanlog");
		out=fopen(f2,"a");
		if (out) {
			fprintf(out,"%d\t%d\t%d\t%s\t%s\n",ref,tm,mode+10,strrchr(file,'/')+1,c);
			fclose(out);
		}
		free(f2);
		free(c);
	}
}

char s[2048];
sprintf(s,"%d\t%d\t%d\t%s\t%s\n",ref,tm,mode,file,text);

#ifdef FILESERVER
if (FSHOST) { 
	FSsend(FS_TEXT,strlen(s)+1, ".emanlog", s);
	return;		// for remote logging, we skip the dirlist 
}
#endif

out=fopen(".emanlog","a");

if (out) {
	file_lock_wait(out);
	fprintf(out,"%s",s);
	file_unlock(out);
	fclose(out);
}

}

void LOGerror(int ref,int err,char *text) {
FILE *out;
time_t tm=time(0);

if (!text) text="";

char s[2048];
sprintf(s,"%d\t%d\t%d\t%d\t%s\n",ref,tm,LOG_ERROR,err,text);

#ifdef FILESERVER
if (FSHOST) { 
	FSsend(FS_TEXT,strlen(s)+1, ".emanlog", s);
	return;		// for remote logging, we skip the dirlist 
}
#endif

out=fopen(".emanlog","a");
if (!out) return;
file_lock_wait(out);

fprintf(out,"%s",s);

file_unlock(out);
fclose(out);
}


// return a random number between lo and hi
float frand(float lo, float hi)
{
static int f=1;

if (f) { srandom(time(0)); f=0; }

return ((float)random()/2147483647.0*(hi-lo)+lo);
}

// interpolate x,y at x=xl, x must increase 
float interpolate(float xl,int n,float *x,float *y) {
int i,st;
st=(int)floor(n*(xl-x[0])/(x[n-1]-x[0]));
if (st>n-2) st=n-2;

// now we set i such that st straddles x[i], x[i+1]
if (x[st]<=xl && x[st+1]>=xl) i=st;
else if (x[st]>xl) { for (i=st; i>=0 && x[i]>xl; i--); }
else { for (i=st; i<n-1 && x[i]<xl; i++) ; i--; }

// outside range, return terminal value
if (x[i]>st) { printf("Range exceeded %f, %f to %f\n",xl,x[i],x[i+1]); return x[0]; }
if (x[i+1]<st) { printf("Range exceeded %f, %f to %f\n",xl,x[i],x[i+1]); return x[n-1]; }

float f=(xl-x[i])/(x[i+1]-x[i]);
return f*y[i+1]+(1-f)*y[i];
}


// this will interpolate a peak location. v3 is considered to be at x=0
// The other points are nearest neighbors
// returns 0 if not well defined
float p5peak(float v1, float v2,float v3, float v4,float v5) {
float r;

if (v2==v4) return 0;

if (v2<v4) {
	r=(2*v4-v5-v3)/(v3+v4-v2-v5);
	if (r<0||r>1.0) return 0;
//printf("%f\n",r);
	return r;
}

r=(2*v2-v1-v3)/(v4+v1-v3-v2);
if (r<-1.0 || r>0) return 0;
//printf("%f\n",r);
return r;
}

float grand(float mean,float sigma)
{
float x,y,r,f;

do {
	x=frand(-1.0,1.0);
	y=frand(-1.0,1.0);
	r=x*x+y*y;
} while (r>1.0||r==0);
f=sqrt(-2.0*log(r)/r);

return x*f*sigma+mean;
}


FILE *DFopenr(const char *name) {
char s[MAXPATHLEN];

if (!name) { printf("Attempt to open null file\n"); return NULL; }
if (!getenv("HOME")) { printf("$HOME not set!\n"); exit(1); }
sprintf(s,"%s/.eman/%s",getenv("HOME"),name);

return fopen(s,"r");
}

FILE *DFopena(const char *name) {
char s[MAXPATHLEN];
FILE *ret;

if (!name) { printf("Attempt to open null file\n"); return NULL; }
if (!getenv("HOME")) { printf("$HOME not set!\n"); exit(1); }
sprintf(s,"%s/.eman",getenv("HOME"));
if (access(s,F_OK)) mkdir(s,0xffff);

sprintf(s,"%s/.eman/%s",getenv("HOME"),name);
ret=fopen(s,"a");

return ret;
}

FILE *DFopenw(const char *name) {
char s[MAXPATHLEN];
FILE *ret;

if (!name) { printf("Attempt to open null file\n"); return NULL; }
if (!getenv("HOME")) { printf("$HOME not set!\n"); exit(1); }
sprintf(s,"%s/.eman",getenv("HOME"));
if (access(s,F_OK)) mkdir(s,0xffff);

sprintf(s,"%s/.eman/%s",getenv("HOME"),name);
ret=fopen(s,"w");
chmod(s,00600);

return ret;
}

char *DFdir() {
char *s = (char *)malloc(MAXPATHLEN);

sprintf(s,"%s/.eman",getenv("HOME"));
if (access(s,F_OK)) mkdir(s,0xffff);

sprintf(s,"%s/.eman",getenv("HOME"));

return s;
}

void roteulertb(float *alt,float *az,float *phi, float dx,float dy)
{
float mx1[3][3],mx2[3][3],mx3[3][3];
float a,t,c,s,x,y;
int i,j,k;

x=dy/hypot(dx,dy);	// axis
y=-dx/hypot(dx,dy);
a=-hypot(dx,dy)/100.0;	// angle
t=1.0-cos(a);
c=cos(a);
s=sin(a);


mx1[0][0]=(cos(*phi)*cos(*az)-cos(*alt)*sin(*az)*sin(*phi));
mx1[0][1]=-(sin(*phi)*cos(*az)+cos(*alt)*sin(*az)*cos(*phi));
mx1[0][2]=sin(*alt)*sin(*az);
mx1[1][0]=(cos(*phi)*sin(*az)+cos(*alt)*cos(*az)*sin(*phi));
mx1[1][1]=(sin(*phi)*sin(*az)+cos(*alt)*cos(*az)*cos(*phi));
mx1[1][2]=-sin(*alt)*cos(*az);
mx1[2][0]=sin(*alt)*sin(*phi);
mx1[2][1]=sin(*alt)*cos(*phi);
mx1[2][2]=cos(*alt);

mx2[0][0]=t*x*x+c;
mx2[0][1]=t*x*y;
mx2[0][2]=-s*y;
mx2[1][0]=t*x*y;
mx2[1][1]=t*y*y+c;
mx2[1][2]=s*x;
mx2[2][0]=s*y;
mx2[2][1]=-s*x;
mx2[2][2]=c;

for (i=0; i<3; i++) {
	for (j=0; j<3; j++) {
		mx3[i][j]=0;
		for (k=0; k<3; k++) {
			mx3[i][j]+=mx1[i][k]*mx2[k][j];
		}
	}
}
	

if (fabs(*alt)<.0001) {
	*phi=0;
	if (fabs(mx3[1][1])>1.0) *az=0;	// not a good thing
	else *az=acos(mx3[1][1]);
}
else {
	*az=atan2(mx3[0][2],-mx3[1][2]);
	*phi=atan2(mx3[2][0],mx3[2][1]);
}

*alt=atan2(sqrt(mx3[2][0]*mx3[2][0]+mx3[2][1]*mx3[2][1]),mx3[2][2]);


if (az!=az || phi!=phi) {
printf("%f %f %f\n",mx1[2][0]*mx2[0][2]+mx1[2][1]*mx2[1][2]+mx1[2][2]*mx2[2][2],
	mx1[2][0]*mx2[0][0]+mx1[2][1]*mx2[1][0]+mx1[2][2]*mx2[2][0],
	mx1[0][0]*mx2[0][2]+mx1[0][1]*mx2[1][2]+mx1[0][2]*mx2[2][2]);
az=phi=0;
}

return;
}

void swapint(char *d,int n)
{
int i;
unsigned char s;

for (i=0; i<n; i++,d+=4) {
	s=d[0]; d[0]=d[3]; d[3]=s;
	s=d[1]; d[1]=d[2]; d[2]=s;
}
}

void swapshort(char *d,int n)
{
int i;
unsigned char s;

for (i=0; i<n; i++,d+=2) {
	s=d[0]; d[0]=d[1]; d[1]=s;
}
}

int npfa(int low) { return low; }

// finds fft sizes with good primes. needs some work ...
int bestfftsize(int low) {
static char *valid = NULL;	//array containing valid sizes <1024 for speed

if (!valid) {
	valid=(char *)calloc(4096,1);
	
	for (float i2=1; i2<12.0; i2+=1.0) {
		for (float i3=0; i3<8.0; i3+=1.0) {
			for (float i5=0; i5<6.0; i5+=1.0) {
				for (float i7=0; i7<5.0; i7+=1.0) {
					float f=pow(2.0,i2)*pow(3.0,i3)*pow(5.0,i5)*pow(7.0,i7);
					if (f>4095.0) continue; 
					int n=(int)f;
					valid[n]=1;
				}
			}
		}
	}
}

for (int i=low; i<4096; i++) if (valid[i]) return i;

error(ERR_EXIT,"Sorry, can only find good fft sizes up to 4096 right now.");
return 1;
}

// The new fft routine can handle any size (although odd will make other things crash)
int isvalidfft(int low) { return 1; }

// the new fft library can do any size, so these are obsolete
/*
int npfa(int low)
{
while (!isvalidfft(low)) low++;
return low;
}

int isvalidfft(int low)
{
int v[] = { 2,3,5,7,11,13,17,19 };
int i;

if (low&1) return 0;
low/=2;
while (low!=1) {
        for (i=0; i<8; i++) {
                if (low%v[i]==0) { low/=v[i]; break; }
        }
        if (i==8) return 0;
}
return 1;
}
*/

// least squares fit returns slope, m, and intercept, b
// if ign0 set, ignores all zero values
void lsqfit(int n,float *dx,float *dy,float *m,float *b,int ign0) {
double S=0,Sx=0,Sy=0,Sxx=0,Sxy=0;
int i;
for (i=0; i<n; i++) {
	if (ign0&&(dx[i]==0||dy[i]==0)) continue;
	double y=dy[i];
	double x;
	if (dx) x=dx[i]; else x=i;
	Sx+=x;
	Sy+=y;
	Sxx+=x*x;
	Sxy+=x*y;
	S++;
}

double del=S*Sxx-Sx*Sx;
if (del==0) del=.0000001;
*b=(Sxx*Sy-Sx*Sxy)/del;
*m=(S*Sxy-Sx*Sy)/del;
}


void save_data(float *x, float *y, int n, const char *fsp)
{
FILE *out;
int i;

out=fopen(fsp,"w");
if (out==NULL) return;
for (i=0; i<n; i++) fprintf(out,"%g\t%g\n",x[i],y[i]);
fclose(out);
return;
}

void save_data(float x0,float dx, float *y, int n, const char *fsp)
{
FILE *out;
int i;

out=fopen(fsp,"w");
if (out==NULL) return;
for (i=0; i<n; i++) fprintf(out,"%g\t%g\n",x0+dx*i,y[i]);
fclose(out);
return;
}

// substring search (broken in early linux libc, and missing on SGI)
char *memmem(char *haystack,int hsize,char *needle,int nsize) {
int i;
for (i=0; i<hsize-nsize+1; i++) {
	if (haystack[i]==needle[0]) {
		int j;
		for (j=1; j<nsize; j++) if (haystack[i+j]!=needle[j]) break;
		if (j==nsize) return haystack+i;
	}
}
return NULL;
}


// basically a version of 'system'
int run(char *s) {
pid_t f=fork();
int stat;

#ifdef ALPHA
if (!f) execl("/bin/zsh","/bin/zsh","-c",s,0);	// temporary fix
#else
if (!f) execl("/bin/sh","/bin/sh","-c",s,0);
#endif
//printf("%d exit\n",waitpid(f,NULL,0));
waitpid(f,&stat,0);

if (WIFEXITED(stat)) return WEXITSTATUS(stat);

return 256;
}

// No shared memory for now
/*
inline void *Smalloc(int size,int shared) {
if (shared) return DMEM.malloc(size);
else return malloc(size);
}

inline void Sfree(void *ptr, int shared) {
if (shared) DMEM.free(ptr);
else free(ptr);
}

inline void *Srealloc(void *ptr, int size, int shared) {
if (shared) return DMEM.realloc(ptr,size);
else return realloc(ptr,size)
}
*/

unsigned long resolv(char *host)
{
struct hostent *he;
unsigned long addr;

if (!host) { 
	host=(char *)malloc(128);
	gethostname(host,127);
	he=gethostbyname(host);
	free(host);
}
else he=gethostbyname(host);
if (he==NULL) return(0);
memcpy(&addr,he->h_addr,4);
return(ntohl(addr));
}

int sock_open(unsigned long addr,unsigned short port)
{
int s,p;
sockaddr_in sa;

s=socket(AF_INET,SOCK_STREAM,0);
if (s==-1) return (-1);

memset((char *)&sa, 0,sizeof(sockaddr_in));
sa.sin_family=AF_INET;
sa.sin_port=htons(port);
sa.sin_addr.s_addr=htonl(addr);
if (connect(s,(sockaddr*)&sa,sizeof(sockaddr_in))) return(-2);

return(s);
}

int writesock(int sock,const char *data,int len) {
int rd = 0,r;

while (rd<len) {
	r=write(sock,data+rd,len-rd);
	if (r<0) { printf("ERROR: No write to socket (%d)\n",errno); return -1; }
	rd+=r;
}

return rd;
}

int readsock(int sock,char *data,int len) {
int rd = 0,r;

while (rd<len) {
	r=read(sock,data+rd,len-rd);
	if (r<0) { error(ERR_CRIT,"No data on socket!\n"); return -1; }
	rd+=r;
}

return rd;
}

#ifdef FILESERVER

unsigned short FSPORT=0;
unsigned int FSHOST=0;
unsigned int FSKEY=0;

// Send a command to the fileserver, if datalen==-1, null terminated string
// command is FS_DONE, FS_TEXT, FS_BIN or FS_IMAGIC
// (for FS_BIN, the first 4 bytes are the file location)
int FSsend(int command, int datalen,const char *fsp,const char *data,int fileno) {
int sock=-1,n=0;

if (!FSHOST) return FSsendlocal(command,datalen,fsp,data,fileno);
char path[MAXPATHLEN];
getcwd(path,MAXPATHLEN-1);

while (sock<0 && n<10) {
	sleep(1);
	sock=sock_open(FSHOST,FSPORT);
	n++;
	if (n>1) {
		char s[120];
		gethostname(s,119);
		printf("%s:Retry socket %d\n",s,n);
	}
}

if (sock<0) { printf("%hu,%x  (%d,%d)\n",FSPORT,FSHOST,sock,errno); error(ERR_EXIT,"No socket to fileserver after 10 tries"); }

FSCOMMAND com;
if (datalen==-1) com.datalen=strlen(data);
else com.datalen=datalen;
com.command=command;
com.key=FSKEY;
com.pathlen=strlen(path);
com.filelen=strlen(fsp);
com.fileno=fileno;
writesock(sock,(char *)&com,sizeof(com));
writesock(sock,(char *)path,strlen(path));
writesock(sock,(char *)fsp,strlen(fsp));
writesock(sock,(char *)data,com.datalen);
close(sock);
}
#else
int FSsend(int command, int datalen,const char *fsp,const char *data,int fileno) {
FSsendlocal(command,datalen,fsp,data,fileno);
}
#endif

int FSsendlocal(int command, int datalen,const char *fsp,const char *data,int fileno) {
FILE *out;
int *loc = (int *) data;

switch(command) {
case FS_BIN:
	if (*loc>=0) {
		out=fopen(fsp,"r+");
		if (!out) out=fopen(fsp,"w");
		file_lock_wait(out);
		int *loc = (int *)data;
		fseek(out,SEEK_SET,*loc);
	}
	else {					// -1 appends, -2 truncates
		if (*loc==-1) out=fopen(fsp,"a");
		else out=fopen(fsp,"w");
		file_lock_wait(out);
	}
	fwrite(data+4,datalen-4,1,out);
	fclose(out);
	break;
case FS_LST:
	if (access(fsp,F_OK)||!fileno) {
		out=fopen(fsp,"w");
		fprintf(out,"#LST\n");
	}
	else out=fopen(fsp,"a");
	file_lock_wait(out);
	fprintf(out,"%s",data);
	fclose(out);
	break;
case FS_TEXT:
	out=fopen(fsp,"a");
	file_lock_wait(out);
	fprintf(out,"%s",data);
	fclose(out);
	break;
}

}

