// make3dicos2.C  Copyright 7/21/00 Steve Ludtke
// This program should only be called from make3dicos. It does the hard work
// of building a 3d icosahedral model in Fourier space. However, it uses VolFrag
// objects, which aren't very useful outside the context of these 2 programs

//N make3dicos2
//F Do not use this program directly. It is called from make3dicos.
//T X
//1
//P <input file>	File containing class averages
//P <outfile>	Output 3D MRC file
//P [pad=<padsize>]	To reduce Fourier artifacts, the model is typically padded by ~25%
//U make3dicos classes.1.img out=threed.1.mrc pad=128

//D This program should only be called from make3dicos. It does the hard work
//E of building a 3d icosahedral model in Fourier space. However, it uses VolFrag
//E objects, which aren't very useful outside the context of these 2 programs


#include <stdio.h>
#include "EMData.h"
#include "List.h"
#include "util.h"
#include "Euler.h"
#include <math.h>
#include <ctype.h>
#include "VolFrag.h"

#define I2G   (4.0/(PI*PI))

main(int argc, char *argv[]) {
int pad=0,n,i,j,num=0,denom=1;
char ind[MAXPATHLEN];
char *file="threed.mrc";

appinit(argc,argv);

if (argc<2) { appusage(argv[0]); exit(1); }

n=fileCount(argv[1]);
if (n<3) { printf("Input file too small ?!?!\n"); exit(1); }

for (i=3; i<argc; i++) {	
	if (strncmp(argv[i],"pad=",4)==0) pad=atoi(&argv[i][4]);
	else if (strncmp(argv[i],"frac=",5)==0) sscanf(&argv[i][5],"%d/%d",&num,&denom);
	else if (strncmp(argv[i],"usefs",5)==0) continue;
	else { printf("Unknown argument '%s'\n",argv[i]); exit(1); }
}

int Ref=LOGbegin(argc,argv);
LOG(Ref,argv[1],LOG_INFILE,NULL);

n=fileCount(argv[1]);
if (n<5) error("Must be >5 particles\n",ERR_EXIT);

// if no pad, set it to size of input images
if (!pad) {
	EMData dr;
	dr.readImage(argv[1],0);
	pad=dr.ySize();
}

// Set up volume fragments for real, imag and norm
VolFrag fragr,fragi,fragn;
fragr.setSize(pad);
fragr.setupIcos();
fragr.zero();
fragi.setSize(pad);
fragi.setupIcos();
fragi.zero();
fragn.setSize(pad);
fragn.setupIcos();
fragn.zero();

int ns=0,c=0;
EMData dr,*dc,*df;
float *data;

for (i=num*n/denom; i<n*(num+1)/denom; i++) {
	// get image, size and transform
	dr.readImage(argv[1],i);
	dc=dr.clip((dr.xSize()-pad)/2,(dr.ySize()-pad)/2,pad,pad);
	dc->toCorner();
	df=dc->doFFT();
	data=df->getDataRO();

	printf("%d. %f\t%f\t%f\n",i,dr.alt()*180.0/PI,dr.az()*180.0/PI,dr.phi()*180.0/PI);
	
	// now we insert 60 copies into frag, this is very similar to
	// insertSlice in EMData
	Euler a=*dr.getEuler(),e;
	a.setSym("icos");
	for (e=a.FirstSym(); e.valid(); e=a.NextSym(),ns++) {
//		printf(" %d\r",ns); fflush(stdout);
		float mx[6],x,y,xx,yy,zz,dx,dy,dz,cc;
		int x0,y0,z0;
		
		mx[0]=(cos(e.phi())*cos(e.az())-cos(e.alt())*sin(e.az())*sin(e.phi()));
		mx[1]=-(sin(e.phi())*cos(e.az())+cos(e.alt())*sin(e.az())*cos(e.phi()));
		mx[2]=(cos(e.phi())*sin(e.az())+cos(e.alt())*cos(e.az())*sin(e.phi()));
		mx[3]=(-sin(e.phi())*sin(e.az())+cos(e.alt())*cos(e.az())*cos(e.phi()));
		mx[4]=sin(e.alt())*sin(e.phi());
		mx[5]=sin(e.alt())*cos(e.phi());


		for (y=-pad/2,j=0; y<pad/2; y++) {
			for (x=0; x<pad/2+1; x++,j+=2) {
				xx=x*mx[0]+y*mx[1];
				yy=x*mx[2]+y*mx[3];
				zz=x*mx[4]+y*mx[5];

				if (xx<0) { xx=-xx; yy=-yy; zz=-zz; cc=-1.0; } else cc=1.0;
				yy+=pad/2;
				zz+=pad/2;

				x0=(int)floor(xx);
				y0=(int)floor(yy);
				z0=(int)floor(zz);

				if (!fragr.validXYZ(x0,y0,z0)) continue;

				dx=xx-x0;
				dy=yy-y0;
				dz=zz-z0;

				float g[8],dr,di;

				dr=data[j];
				di=data[j+1]*cc;

				g[0]=agauss(1.0,dx,dy,dz,I2G);
				g[1]=agauss(1.0,1.0-dx,dy,dz,I2G);
				g[2]=agauss(1.0,dx,1.0-dy,dz,I2G);
				g[3]=agauss(1.0,1.0-dx,1.0-dy,dz,I2G);
				g[4]=agauss(1.0,dx,dy,1.0-dz,I2G);
				g[5]=agauss(1.0,1.0-dx,dy,1.0-dz,I2G);
				g[6]=agauss(1.0,dx,1.0-dy,1.0-dz,I2G);
				g[7]=agauss(1.0,1.0-dx,1.0-dy,1.0-dz,I2G);

				fragr.addValue(x0,y0,z0,g[0]*dr);
				fragi.addValue(x0,y0,z0,g[0]*di);
				fragn.addValue(x0,y0,z0,g[0]);

				fragr.addValue(x0+1,y0,z0,g[1]*dr);
				fragi.addValue(x0+1,y0,z0,g[1]*di);
				fragn.addValue(x0+1,y0,z0,g[1]);

				fragr.addValue(x0,y0+1,z0,g[2]*dr);
				fragi.addValue(x0,y0+1,z0,g[2]*di);
				fragn.addValue(x0,y0+1,z0,g[2]);

				fragr.addValue(x0+1,y0+1,z0,g[3]*dr);
				fragi.addValue(x0+1,y0+1,z0,g[3]*di);
				fragn.addValue(x0+1,y0+1,z0,g[3]);

				fragr.addValue(x0,y0,z0+1,g[4]*dr);
				fragi.addValue(x0,y0,z0+1,g[4]*di);
				fragn.addValue(x0,y0,z0+1,g[4]);

				fragr.addValue(x0+1,y0,z0+1,g[5]*dr);
				fragi.addValue(x0+1,y0,z0+1,g[5]*di);
				fragn.addValue(x0+1,y0,z0+1,g[5]);

				fragr.addValue(x0,y0+1,z0+1,g[6]*dr);
				fragi.addValue(x0,y0+1,z0+1,g[6]*di);
				fragn.addValue(x0,y0+1,z0+1,g[6]);

				fragr.addValue(x0+1,y0+1,z0+1,g[7]*dr);
				fragi.addValue(x0+1,y0+1,z0+1,g[7]*di);
				fragn.addValue(x0+1,y0+1,z0+1,g[7]);
			}
		}
	}

	df->doneData();
	delete dc;	// this gets df too
}

sprintf(ind,"%s.r.frg",argv[2]);
fragr.writeFile(ind);

sprintf(ind,"%s.i.frg",argv[2]);
fragi.writeFile(ind);

sprintf(ind,"%s.n.frg",argv[2]);
fragn.writeFile(ind);

printf("c=%d\n",c);
LOG(Ref,NULL,LOG_END,NULL);

}
